package cn.kdan.pdf.tech.core.service.impl;

import cn.kdan.pdf.tech.core.constant.VppDeviceConstant;
import cn.kdan.pdf.tech.core.enums.CaptchaActionEnum;
import cn.kdan.pdf.tech.core.enums.DeviceStatusEnum;
import cn.kdan.pdf.tech.core.enums.LicenseCodeStatusEnum;
import cn.kdan.pdf.tech.core.mapper.ext.ExtVppDeviceMapper;
import cn.kdan.pdf.tech.core.model.*;
import cn.kdan.pdf.tech.core.pojo.vo.VppDeviceVO;
import cn.kdan.pdf.tech.core.pojo.vo.VppRTeamMemberRoleVO;
import cn.kdan.pdf.tech.core.service.*;
import cn.kdan.pdf.tech.core.utils.ThreadPoolSingleUtil;
import com.github.pagehelper.PageHelper;
import com.github.pagehelper.PageInfo;
import constant.CommonConstant;
import exception.BackendRuntimeException;
import lombok.extern.slf4j.Slf4j;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.util.CollectionUtils;
import org.springframework.util.ObjectUtils;
import cn.kdan.pdf.tech.core.utils.EmailUtils;

import javax.annotation.Resource;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Date;
import java.util.List;
import java.util.stream.Collectors;


/**
 * 用户与团队的关系类
 *
 * @author tangxiangan
 */
@Slf4j
@Service
public class VppDeviceServiceImpl implements VppDeviceService {

    @Resource
    private ExtVppDeviceMapper extVppDeviceMapper;
    @Resource
    private VppMemberService memberService;
    @Resource
    private ProductService productService;
    @Autowired
    private EmailUtils emailUtils;
    @Resource
    private SubscriptionService subscriptionService;
    @Resource
    private VppRTeamMemberRoleService vppRTeamMemberRoleService;
    @Resource
    private VppLicenseCodeService licenseCodeService;

    @Override
    public PageInfo<VppDeviceVO> page(Integer page, Integer pageSize, Integer productId, Integer status, String teamId, String queryString) {
        VppMember member = memberService.getCurrentUser();
        List<String> teamIds = new ArrayList<>();
        if(ObjectUtils.isEmpty(teamId)){
            List<VppRTeamMemberRoleVO> list = vppRTeamMemberRoleService.listByMemberIdAdmin(member.getId());
            List<String> idList  = list.stream().map(VppRTeamMemberRoleVO::getTeamId).collect(Collectors.toList());
            teamIds = idList;
        }else{
            teamIds.add(teamId);
        }
        if(CollectionUtils.isEmpty(teamIds)){
            throw new BackendRuntimeException(VppDeviceConstant.EXCEPTION_MSG_NO_TEAM);
        }
        PageHelper.startPage(page, pageSize);
        return new PageInfo<>(extVppDeviceMapper.listWithInfo(teamIds, productId, status, queryString));
    }

    @Override
    public List<VppDeviceVO> listWithInfo(Integer productId, Integer status, String teamId, String queryString) {
        VppMember member = memberService.getCurrentUser();
        memberService.getCurrentUser();
        List<String> teamIds = new ArrayList<>();
        if(ObjectUtils.isEmpty(teamId)){
            List<VppRTeamMemberRoleVO> list = vppRTeamMemberRoleService.listByMemberIdAdmin(member.getId());
            List<String> idList  = list.stream().map(VppRTeamMemberRoleVO::getTeamId).collect(Collectors.toList());
            teamIds = idList;
        }else{
            teamIds.add(teamId);
        }
        if(CollectionUtils.isEmpty(teamIds)){
            throw new BackendRuntimeException(VppDeviceConstant.EXCEPTION_MSG_NO_TEAM);
        }
        return extVppDeviceMapper.listWithInfo(teamIds, productId, status, queryString);
    }


    @Override
    public void unBindByMemberId(String memberId, List<String> cdKeys) {
        //查出用户所有已经激活的设备
        DevicesExample example = new DevicesExample();
        DevicesExample.Criteria criteria = example.createCriteria().andStatusEqualTo(DeviceStatusEnum.ACTIVE.code());
        if (!CollectionUtils.isEmpty(cdKeys)) {
            criteria.andCdkeyIn(cdKeys);
        }
        //修改设备为未授权
        Devices devices = new Devices();
        devices.setStatus(DeviceStatusEnum.UNACTIVATED.code());
        devices.setSubscriptionId(null);
        devices.setEndUpAt(null);
        devices.setCanceledDate(new Date());
        extVppDeviceMapper.updateByExampleSelective(devices, example);
    }

    @Transactional
    @Override
    public void cancel(String deviceId) {
        Devices devices = extVppDeviceMapper.selectByPrimaryKey(deviceId);
        if (ObjectUtils.isEmpty(devices)) {
            throw new BackendRuntimeException(VppDeviceConstant.EXCEPTION_MSG_DEVICE_NOT_EXIST);
        }

        //设置序列码状态
        LicenseCodes code = licenseCodeService.getByCdKey(devices.getCdkey());
        VppMember vppMember = memberService.getById(code.getVppMemberId());
        log.info("send cancel email!");
        ThreadPoolSingleUtil.getInstance().excute(() -> {
            if(!ObjectUtils.isEmpty(vppMember)){
                Subscriptions subscriptions = subscriptionService.getById(code.getSubscriptionId());
                Products product = productService.getById(subscriptions.getProductId());
                String adminEmail =  memberService.getCompanySuperAdmin(vppMember.getCompanyId());
                String text = String.format(CaptchaActionEnum.CANCEL_DEVICE_ASSIGN.text(),vppMember.getEmail(),devices.getModel(),product.getName(),product.getName(),devices.getModel(),adminEmail) + CommonConstant.signHtml;
                emailUtils.sendMail(CaptchaActionEnum.CANCEL_DEVICE_ASSIGN.subject(), text, Collections.singletonList(vppMember.getEmail()));
            }
        });
        //更新设备状态
        devices.setStatus(DeviceStatusEnum.UNACTIVATED.code());
        devices.setCanceledDate(new Date());
        devices.setSubscriptionId(null);
        devices.setEndUpAt(null);
        devices.setCdkey(null);
        log.info("update device:" + devices);
        extVppDeviceMapper.updateByPrimaryKeySelective(devices);
        code.setUsedTimes((short) 0);
        code.setValidFlag(LicenseCodeStatusEnum.ASSIGNED.code());
        code.setUpdatedAt(new Date());
        log.info("update code:" + code);
        licenseCodeService.update(code);
    }

    /**
     * 获取序列激活的设备
     * @param cdKeyList 序列码
     * @return
     */
    @Override
    public List<Devices> getByCdKeyList(List<String> cdKeyList) {
        DevicesExample devicesExample = new DevicesExample();
        devicesExample.createCriteria().andCdkeyIn(cdKeyList).andStatusEqualTo(1);
        return extVppDeviceMapper.selectByExample(devicesExample);
    }
}
