package cn.kdan.pdf.tech.core.service.impl;

import cn.hutool.core.date.DateUtil;
import cn.kdan.pdf.tech.core.constant.AuthConstant;
import cn.kdan.pdf.tech.core.constant.VppDeviceConstant;
import cn.kdan.pdf.tech.core.constant.VppLicenseCodeConstant;
import cn.kdan.pdf.tech.core.constant.VppMemberConstant;
import cn.kdan.pdf.tech.core.enums.CaptchaActionEnum;
import cn.kdan.pdf.tech.core.enums.LicenseCodeStatusEnum;
import cn.kdan.pdf.tech.core.mapper.ext.ExtVppLicenseCodeMapper;
import cn.kdan.pdf.tech.core.model.*;
import cn.kdan.pdf.tech.core.pojo.vo.VppLicenseCodeVO;
import cn.kdan.pdf.tech.core.pojo.vo.VppRTeamMemberRoleVO;
import cn.kdan.pdf.tech.core.service.*;
import cn.kdan.pdf.tech.core.utils.ExcelUtils;
import cn.kdan.pdf.tech.core.utils.FileUtil;
import cn.kdan.pdf.tech.core.utils.ThreadPoolSingleUtil;
import com.github.pagehelper.PageHelper;
import com.github.pagehelper.PageInfo;
import com.google.common.util.concurrent.ThreadFactoryBuilder;
import constant.CommonConstant;
import exception.BackendRuntimeException;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang.StringUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.util.CollectionUtils;
import org.springframework.util.ObjectUtils;
import org.springframework.web.multipart.MultipartFile;
import cn.kdan.pdf.tech.core.utils.EmailUtils;

import javax.annotation.Resource;
import java.io.IOException;
import java.io.InputStream;
import java.text.MessageFormat;
import java.util.*;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;
import java.util.stream.Collectors;


/**
 * 序列码
 *
 * @author tangxiangan
 */
@Slf4j
@Service
public class VppLicenseCodeServiceImpl implements VppLicenseCodeService {

    @Resource
    private ExtVppLicenseCodeMapper licenseCodeMapper;
    @Resource
    private VppDeviceService vppDeviceService;
    @Resource
    private VppTeamService vppTeamService;
    @Resource
    private VppMemberService vppMemberService;
    @Resource
    private VppRTeamMemberRoleService vppRTeamMemberRoleService;
    @Autowired
    private EmailUtils emailUtils;
    @Autowired
    private ProductService productService;
    @Value("${ssoUrl}")
    private String ssoUrl;
    private static final ThreadPoolExecutor LICENSE_CODE_SEND_MAIL_THREAD_POOL = new ThreadPoolExecutor( Runtime.getRuntime().availableProcessors(),
            Runtime.getRuntime().availableProcessors(),
            1L,
            TimeUnit.MINUTES,
            new LinkedBlockingQueue<>(300), new ThreadFactoryBuilder()
            .setNameFormat("LICENSE_CODE_SEND_MAIL_THREAD_POOL").build(), new ThreadPoolExecutor.AbortPolicy());


    @Override
    public void unBindByMemberId(String memberId, List<String> teamIds) {
        LicenseCodesExample example = new LicenseCodesExample();
        LicenseCodesExample.Criteria criteria = example.createCriteria().andVppMemberIdEqualTo(memberId);
        //如果是团队管理员，他可能会在多个团队
        if (!CollectionUtils.isEmpty(teamIds)) {
            criteria.andTeamIdIn(teamIds);
        }
        List<LicenseCodes> list = licenseCodeMapper.selectByExample(example);
        LicenseCodes vppLicenseCode = new LicenseCodes();
        vppLicenseCode.setVppMemberId("");
        vppLicenseCode.setUsedTimes((short) 0);
        vppLicenseCode.setValidFlag(LicenseCodeStatusEnum.CANCELED.code());
        vppLicenseCode.setUpdatedAt(new Date());
        if (!CollectionUtils.isEmpty(list)) {
            //用户id改为空，序列码状态变为取消授权
            licenseCodeMapper.updateByExampleSelective(vppLicenseCode, example);
            List<String> cdKeys = list.stream().map(LicenseCodes::getCdkey).collect(Collectors.toList());
            //解绑序列码绑定的所有设备
            vppDeviceService.unBindByMemberId(memberId, cdKeys);
            //归还团队可用序列码数量
            list.forEach(item -> {
                reduceLicenseRemainNum(list.size(), item.getTeamId());
            });
        }
    }

    private void reduceLicenseRemainNum(Integer num, String teamId) {
        VppTeam vppTeam = vppTeamService.getById(teamId);
        int licenseRemain = vppTeam.getLicenseRemain() + num;
        vppTeam.setLicenseRemain(licenseRemain);
        vppTeamService.updateTeam(vppTeam);
    }

    @Override
    public PageInfo<VppLicenseCodeVO> page(Integer page, Integer pageSize, Integer productId, String teamId,
                                           Integer status, String queryString) {
        VppMember member = vppMemberService.getCurrentUser();
        List<String> teamIds = new ArrayList<>();
        handleTeamIds(teamId, member, teamIds);
        if(CollectionUtils.isEmpty(teamIds)){
            throw new BackendRuntimeException(VppDeviceConstant.EXCEPTION_MSG_NO_TEAM);
        }
        PageHelper.startPage(page, pageSize);
        return new PageInfo<>(licenseCodeMapper.listWithInfo(productId, teamIds, status, queryString));
    }

    @Override
    public List<VppLicenseCodeVO> listWithInfo(Integer productId, String teamId, Integer status, String queryString) {
        VppMember member = vppMemberService.getCurrentUser();
        List<String> teamIds = new ArrayList<>();
        handleTeamIds(teamId, member, teamIds);
        return licenseCodeMapper.listWithInfo(productId, teamIds, status, queryString);
    }

    private void handleTeamIds(String teamId, VppMember member, List<String> teamIds) {
        if(ObjectUtils.isEmpty(teamId)){
            List<VppRTeamMemberRoleVO> list = vppRTeamMemberRoleService.listByMemberIdAdmin(member.getId());
            List<String> idList  = list.stream().map(VppRTeamMemberRoleVO::getTeamId).collect(Collectors.toList());
            for (String s : idList) {
                teamIds.add(s);
            }
        }else{
            teamIds.add(teamId);
        }
    }

    @Transactional
    @Override
    public void singleAssign(List<String> memberIds, String teamId, List<Integer> productIds) {
        //根据团队id查出团队对应每种产品的剩余未授权序列码数量
        LicenseCodesExample example = new LicenseCodesExample();
        List<Integer> statusList = new ArrayList<>();
        statusList.add(LicenseCodeStatusEnum.UNASSIGNED.code());
        statusList.add(LicenseCodeStatusEnum.CANCELED.code());
        example.createCriteria().andTeamIdEqualTo(teamId).andValidFlagIn(statusList);
        List<LicenseCodes> vppLicenseCodes = licenseCodeMapper.selectByExample(example);
        Map<Integer, List<LicenseCodes>> codeMap = vppLicenseCodes.stream().collect(Collectors.groupingBy(LicenseCodes::getProductId));
        //对比用户数量判断每种产品序列码是否足够
        productIds.forEach(productId -> {
            if (CollectionUtils.isEmpty(codeMap)||!codeMap.containsKey(productId) ||(codeMap.containsKey(productId) && codeMap.get(productId).size() < memberIds.size())) {
                //Products products = productService.getById(productId);
                throw new BackendRuntimeException(VppLicenseCodeConstant.EXCEPTION_MSG_CODE_NOT_ENOUGH_TEAM);
            }
        });
        //对应用户id被授权的序列码记录在这个map，用于后续发邮件通知
        Map<String,List<LicenseCodes>> memberCodeMap = new HashMap<>();
        //如果都足够则开始授权，把对应的产品的序列码分配给用户
        productIds.forEach( item -> {
            List<LicenseCodes> codes = codeMap.get(item);
            for (int i = 0; i < memberIds.size(); i++) {
                LicenseCodes code = codes.get(i);
                code.setVppMemberId(memberIds.get(i));
                code.setValidFlag(LicenseCodeStatusEnum.ASSIGNED.code());
                code.setUpdatedAt(new Date());
                code.setAssignedDate(new Date());
                licenseCodeMapper.updateByPrimaryKeySelective(code);
                if(CollectionUtils.isEmpty(memberCodeMap.get(memberIds.get(i)))){
                    List<LicenseCodes> list = new ArrayList<>();
                    list.add(code);
                    memberCodeMap.put(memberIds.get(i),list);
                }else {
                    memberCodeMap.get(memberIds.get(i)).add(code);
                }
            }
        });
        //修改团队的剩余序列码数量
        reduceLicenseRemainNum(memberIds.size() * productIds.size(), teamId);
        LICENSE_CODE_SEND_MAIL_THREAD_POOL.submit(() -> {
            //通知用户获得了哪个产品的序列码的授权
            memberCodeMap.forEach((key,value) -> {
                List<LicenseCodes> codes = value;
                String text = "";
                String appUrl = "";
                String userWay = "";
                List<String> havingUseWayList = new ArrayList<>();
                for (LicenseCodes code : codes) {
                    Products product = productService.getById(code.getProductId());
                    text = text  + String.format(CommonConstant.codeInfoTemplate,code.getCdkey(),product.getName());
                    appUrl = StringUtils.isEmpty(appUrl) ? appUrl + getAdminLink(product.getCode()) : appUrl;
                    //如果不是同一个产品的序列码则添加使用方式
                    if(!havingUseWayList.contains(product.getCode())){
                        userWay = userWay + getProductUseWayTemplate(product.getCode(),product.getName());
                        havingUseWayList.add(product.getCode());
                    }
                }
                if(userWay.contains("dont")){
                    userWay = userWay.replaceAll("dont","don't");
                }
                if(userWay.contains("cant")){
                    userWay =  userWay.replaceAll("cant","can't");
                }
                VppMember vppMember = vppMemberService.getById(key);
                //用户邮箱 序列码 对应产品主页 公司超管邮箱
                String content = String.format(CaptchaActionEnum.ASSIGN_LICENSE_CODE.text(),vppMember.getEmail(),
                        text,appUrl,vppMemberService.getCompanySuperAdmin(vppMember.getCompanyId()),userWay) + CommonConstant.signHtml;
                emailUtils.sendMail(CaptchaActionEnum.ASSIGN_LICENSE_CODE.subject(), content, Collections.singletonList(vppMember.getEmail()));
            });
        });

    }

    /**
     * 获取不同产品官网链接
     *
     * @param productCode 产品code
     * @return String
     */
    private String getAdminLink (String productCode){
        if(productCode.contains("pdfreaderpro")){
           return "<a href=\"https://www.pdfreaderpro.com/\" target=\"_blank\">HERE</a>";
        }
        return "<a href=\"https://www.pdfreaderpro.com/\" target=\"_blank\">HERE</a>";
    }

    /**
     * 获取不同产品的使用模版
     *
     * @param productCode 产品code
     * @return String
     */
    private String getProductUseWayTemplate (String productCode,String productName){
        if(productCode.contains("pdfreaderpro.mac")){
            return MessageFormat.format(CommonConstant.pdfReaderProMac, productName,ssoUrl, ssoUrl);
        }
        if(productCode.contains("pdfreaderpro.windows")){
            return MessageFormat.format(CommonConstant.pdfReaderProWindows, productName,ssoUrl, ssoUrl);
        }
        return "";
    }


    @Override
    public void teamAssign(String teamId, List<Integer> productIds) {
        //查询团队下有哪些用户
        List<VppRTeamMemberRole> memberVOList = vppRTeamMemberRoleService.listByTeamIdWithoutSuperAdmin(teamId);
        List<String> memberIds = memberVOList.stream().map(VppRTeamMemberRole::getMemberId).collect(Collectors.toList());
        singleAssign(memberIds, teamId, productIds);
    }

    @Override
    public void cancelLicenceAssign(String licenceId) {
        // 获取当前登录用户
        VppMember user = vppMemberService.getCurrentUser();
        LicenseCodes vppLicenseCode = licenseCodeMapper.selectByPrimaryKey(licenceId);
        if (null == vppLicenseCode) {
            throw new BackendRuntimeException(AuthConstant.EXCEPTION_MSG_TOKEN_IS_INVALID);
        }
        // 验证是否为本公司序列码
        if (!vppLicenseCode.getCompanyId().equals(user.getCompanyId())) {
            throw new BackendRuntimeException(VppLicenseCodeConstant.INSUFFICIENT_PERMISSIONS);
        }
        if (vppLicenseCode.getValidFlag() != 2 && vppLicenseCode.getValidFlag() != 3 && vppLicenseCode.getValidFlag() != 8) {
            throw new BackendRuntimeException(VppLicenseCodeConstant.LICENSE_STATUS_ERROR);
        }
        if (StringUtils.isEmpty(vppLicenseCode.getVppMemberId()) || null == vppLicenseCode.getProductId()) {
            throw new BackendRuntimeException(VppLicenseCodeConstant.LICENSE_STATUS_ERROR);
        }
        // 获取被授权人和产品信息
        VppMember vppMember = vppMemberService.getById(vppLicenseCode.getVppMemberId());
        Products products = productService.getById(vppLicenseCode.getProductId());
        vppLicenseCode.setMemberId(null);
        vppLicenseCode.setVppMemberId(null);
        vppLicenseCode.setValidFlag(4);
        vppLicenseCode.setUpdatedAt(new Date());
        vppLicenseCode.setUsedTimes(Short.valueOf("0"));
        licenseCodeMapper.updateByPrimaryKey(vppLicenseCode);
        // 解绑激活设备
        vppDeviceService.unBindByMemberId(null, Collections.singletonList(vppLicenseCode.getCdkey()));
        // 发送邮件
        ThreadPoolSingleUtil.getInstance().excute(() -> {
            String text = String.format(CaptchaActionEnum.CANCEL_LICENSE_ASSIGN.text(), vppMember.getEmail(), products.getName(), products.getName(), user.getEmail()) + CommonConstant.signHtml;
            emailUtils.sendMail(CaptchaActionEnum.CANCEL_LICENSE_ASSIGN.subject(), text, Collections.singletonList(vppMember.getEmail()));
        });
    }

    @Override
    public String cancelLicenceAssignBatch(MultipartFile file) {
        log.info("开始批量取消{}", DateUtil.date().toString("yyyy-MM-dd HH:mm:ss:SSS"));
        // 判断文件类型是否正确
        if (!FileUtil.isExcel(file)) {
            throw new BackendRuntimeException(VppMemberConstant.FILE_ERROR);
        }
        //判断Excel版本
        boolean isOldExcel = FileUtil.isExcelOld(file);
        int count = 0;
        try {
            // 读取数据
            InputStream in = file.getInputStream();
            List<List<String>> dataList = ExcelUtils.read(in, isOldExcel);
            if (CollectionUtils.isEmpty(dataList) || dataList.size() == 1) {
                return "Success:0";
            }
            if (dataList.get(0).size() != 2) {
                log.error("读取批量导入团队成员数据错误，读取到的列数：{}", dataList.get(0).size());
                throw new BackendRuntimeException(VppMemberConstant.FILE_ERROR);
            }
            log.info("导入数据：{}",dataList.toString());
            // 获取当前登录用户
            VppMember user = vppMemberService.getCurrentUser();
            log.info("获取当前登录用户后{}", DateUtil.date().toString("yyyy-MM-dd HH:mm:ss:SSS"));
            // 从第二行开始读取
            for (int i = 1; i < dataList.size(); i++) {
                List<String> rowDate = dataList.get(i);
                if (CollectionUtils.isEmpty(rowDate)) {
                    continue;
                }
                String productName = rowDate.get(0);
                if (StringUtils.isEmpty(productName)) {
                    continue;
                }
                log.info("查询产品{}；{}", productName.trim(), DateUtil.date().toString("yyyy-MM-dd HH:mm:ss:SSS"));
                // 查询产品
                Products product = productService.getByName(productName.trim());
                if (null == product) {
                    continue;
                }
                String email = rowDate.get(1);
                if (StringUtils.isEmpty(email)) {
                    continue;
                }
                // 查询用户，验证用户是否为本公司
                log.info("查询用户{}；{}", email.trim(), DateUtil.date().toString("yyyy-MM-dd HH:mm:ss:SSS"));
                VppMember vppMember = vppMemberService.getByAccount(email.trim());
                if (null == vppMember || !vppMember.getCompanyId().equals(user.getCompanyId())) {
                    continue;
                }
                // 根据产品和用户查询在当前有效期下的已授权序列码数据
                LicenseCodesExample example = new LicenseCodesExample();
                example.createCriteria().andProductIdEqualTo(product.getId()).andVppMemberIdEqualTo(vppMember.getId());
                List<LicenseCodes> vppLicenseCodeList = licenseCodeMapper.selectByExample(example);
                if (CollectionUtils.isEmpty(vppLicenseCodeList)) {
                    continue;
                }
                List<LicenseCodes> collect = vppLicenseCodeList.stream().filter(c -> c.getValidFlag() != 0 &&
                        (null == c.getEndUpAt() || c.getEndUpAt().compareTo(new Date()) > 0)).collect(Collectors.toList());
                if (CollectionUtils.isEmpty(collect)) {
                    continue;
                }
                for (LicenseCodes vppLicenseCode : collect) {
                    if (null == vppLicenseCode) {
                        continue;
                    }
                    vppLicenseCode.setMemberId(null);
                    vppLicenseCode.setVppMemberId(null);
                    vppLicenseCode.setUsedTimes(Short.valueOf("0"));
                    vppLicenseCode.setValidFlag(4);
                    vppLicenseCode.setUpdatedAt(new Date());
                    log.info("取消授权{}；{}", vppLicenseCode, DateUtil.date().toString("yyyy-MM-dd HH:mm:ss:SSS"));
                    licenseCodeMapper.updateByPrimaryKey(vppLicenseCode);
                    // 解绑激活设备
                    log.info("解绑激活设备{}；{}", vppLicenseCode.getCdkey(), DateUtil.date().toString("yyyy-MM-dd HH:mm:ss:SSS"));
                    vppDeviceService.unBindByMemberId(null, Collections.singletonList(vppLicenseCode.getCdkey()));
                    count++;
                    log.info("发送邮件前{}", DateUtil.date().toString("yyyy-MM-dd HH:mm:ss:SSS"));
                    // 发送邮件
                    ThreadPoolSingleUtil.getInstance().excute(() -> {
                        String text = String.format(CaptchaActionEnum.CANCEL_LICENSE_ASSIGN.text(), vppMember.getEmail(), product.getName(), product.getName(), user.getEmail()) + CommonConstant.signHtml;
                        emailUtils.sendMail(CaptchaActionEnum.CANCEL_LICENSE_ASSIGN.subject(), text, Collections.singletonList(vppMember.getEmail()));
                    });
                }
            }
        } catch (IOException e) {
            log.error("批量取消授权码读取数据失败，{}", e.getMessage());
        }
        return "Success:" + count;
    }

    @Override
    public void update(LicenseCodes licenseCode) {
        licenseCodeMapper.updateByPrimaryKeySelective(licenseCode);
    }

    @Override
    public void updateByPrimaryKey(LicenseCodes licenseCode) {
        licenseCodeMapper.updateByPrimaryKey(licenseCode);
    }

    /**
     * 获取公司所有序列码
     * @param companyId 公司id
     * @return
     */
    @Override
    public List<LicenseCodes> getByCompanyId(String companyId) {
        LicenseCodesExample example = new LicenseCodesExample();
        example.createCriteria().andCompanyIdEqualTo(companyId)
                .andValidFlagNotEqualTo(LicenseCodeStatusEnum.DELETED.code());
        return licenseCodeMapper.selectByExample(example);
    }

    @Override
    public List<LicenseCodes> getByCompanyIdRemoveExpired(String companyId) {
        LicenseCodesExample example = new LicenseCodesExample();
        example.createCriteria().andCompanyIdEqualTo(companyId)
                .andValidFlagNotEqualTo(LicenseCodeStatusEnum.DELETED.code()).andValidFlagNotEqualTo(LicenseCodeStatusEnum.EXPIRED.code());
        return licenseCodeMapper.selectByExample(example);
    }

    @Override
    public List<LicenseCodes> getByCompanyAndTeamId(String companyId, String teamId) {
        LicenseCodesExample example = new LicenseCodesExample();
        example.createCriteria().andCompanyIdEqualTo(companyId)
                .andTeamIdEqualTo(teamId);
        return licenseCodeMapper.selectByExample(example);
    }

    /**
     * 获取团队下的所有已授权序列码
     * @param teamId 团队id
     * @return
     */
    @Override
    public List<LicenseCodes> getByTeamId(String teamId) {
        LicenseCodesExample example = new LicenseCodesExample();
        List<Integer> values = new ArrayList<>();
        values.add(LicenseCodeStatusEnum.ASSIGNED.code());
        values.add(LicenseCodeStatusEnum.ACTIVE.code());
        example.createCriteria().andTeamIdEqualTo(teamId).andValidFlagIn(values);
        return licenseCodeMapper.selectByExample(example);
    }

    @Override
    public List<LicenseCodes> getByTeamIdAndProduceId(String companyId, String teamId, Integer productId) {
        LicenseCodesExample example = new LicenseCodesExample();
        example.createCriteria().andCompanyIdEqualTo(companyId)
                .andTeamIdEqualTo(teamId).andProductIdEqualTo(productId);
        return licenseCodeMapper.selectByExample(example);
    }

    @Override
    public List<LicenseCodes> getByTeamIds(String companyId, List<String> teamIdList) {
        LicenseCodesExample example = new LicenseCodesExample();
        example.createCriteria().andCompanyIdEqualTo(companyId).andTeamIdIn(teamIdList);
        return licenseCodeMapper.selectByExample(example);
    }

    @Override
    public List<LicenseCodes> getByTeamIdsAndProduceIds(String companyId, List<String> teamIdList, List<Integer> productIdList) {
        LicenseCodesExample example = new LicenseCodesExample();
        example.createCriteria().andCompanyIdEqualTo(companyId)
                .andTeamIdIn(teamIdList).andProductIdIn(productIdList);
        return licenseCodeMapper.selectByExample(example);
    }

    @Override
    public List<LicenseCodes> getByProduceIds(String companyId, List<Integer> productIdList) {
        LicenseCodesExample example = new LicenseCodesExample();
        LicenseCodesExample.Criteria criteria = example.createCriteria();
        criteria.andCompanyIdEqualTo(companyId);
        if (!CollectionUtils.isEmpty(productIdList)) {
            criteria.andProductIdIn(productIdList);
        }
        return licenseCodeMapper.selectByExample(example);
    }

    @Override
    public List<LicenseCodes> getBySubscriptionIds(List<String> subscriptionIds) {
        return licenseCodeMapper.listBySubscriptionIds(subscriptionIds);
    }


    @Override
    public LicenseCodes getByCdKey(String cdKey) {
        LicenseCodesExample example = new LicenseCodesExample();
        LicenseCodesExample.Criteria criteria = example.createCriteria();
        criteria.andCdkeyEqualTo(cdKey);
        List<LicenseCodes> list = licenseCodeMapper.selectByExample(example);
        return CollectionUtils.isEmpty(list) ?  null: list.get(0) ;
    }

    @Override
    public PageInfo<VppLicenseCodeVO> pageForReseller(Integer page, Integer pageSize, Integer productId, Integer status, String queryString) {
        VppMember member = vppMemberService.getCurrentUser();
        PageHelper.startPage(page, pageSize);
        return new PageInfo<>(licenseCodeMapper.listWithInfoForReseller(productId, status, queryString, member.getCompanyId()));
    }

    @Transactional
    @Override
    public void disable(String licenseCodeId) {
        //禁用序列码
        LicenseCodes code = licenseCodeMapper.selectByPrimaryKey(licenseCodeId);
        code.setId(licenseCodeId);
        code.setUsedTimes(code.getTimes());
        code.setValidFlag(LicenseCodeStatusEnum.Refunded.code());
        code.setUpdatedAt(new Date());
        update(code);
        // 解绑激活设备
        vppDeviceService.unBindByMemberId(null, Collections.singletonList(code.getCdkey()));
    }
}
